import argparse
from daman import daman as dm


def clear_command():
    parser = argparse.ArgumentParser(description="Clears disc space in local storage.")
    parser.add_argument(
        "--space",
        type=int,
        help="Amount of disc space in MegaBytes to clear.",
        required=True,
    )
    parser.add_argument(
        "--ignore-persist",
        action="store_true",
        help="When provided, deletes the requested file on the cloud service as well.",
    )
    args = parser.parse_args()

    dm.clear_disc(space=args.space, force=args.ignore_persist)



from typing import Union
from pathlib import Path
from configparser import ConfigParser

from daman.configure.aws import configure_aws


HOME_DIR = Path().home()
CONFIG_DIR = HOME_DIR / ".daman"
CLOUD_SERVICES = {"aws": configure_aws}


def configure(
    storage_name: str,
    service: str,
    local_dir: Union[Path, str] = None,
    allocated_space: int = None,
    service_settings: dict = None,
) -> None:
    """Short summary.

    Parameters
    ----------
    storage_name : str
        Description of parameter `storage_name`.
    service : str
        Description of parameter `service`.
    local_dir : Union[Path, str]
        Description of parameter `local_dir`.
    allocated_space : int
        Description of parameter `allocated_space`.
    service_settings : dict
        Description of parameter `service_settings`.

    Returns
    -------
    None
        Description of returned object.

    """

    dm_config = ConfigParser(allow_no_value=True)
    # storing service config
    dm_config.add_section("service")
    dm_config["service"] = {"service": service, "name": storage_name}
    # storing local config
    dm_config.add_section("local")
    if local_dir is None:
        local_dir = str((CONFIG_DIR / "data").resolve())

    dm_config["local"] = {
        "data_dir": str(Path(local_dir).resolve()),
        "registery": str(CONFIG_DIR / "registery.json"),
    }

    if allocated_space is not None:
        dm_config["local"]["allocated_space"] = allocated_space

    CONFIG_DIR.mkdir(parents=True, exist_ok=True)
    with (CONFIG_DIR / "config").open("w") as fw:
        dm_config.write(fw)

    if service_settings is not None:
        configure_service(service=service, settings=service_settings)


def configure_service(service: str, settings: dict) -> None:
    """Short summary.

    Parameters
    ----------
    service : str
        Description of parameter `service`.
    settings : dict
        Description of parameter `settings`.

    Returns
    -------
    None
        Description of returned object.

    """
    service = service.lower()

    # check service is supported
    assert_msg = f"'{service}' is not currently supported. Available services: {list(CLOUD_SERVICES.keys())}"
    assert service in CLOUD_SERVICES, assert_msg

    # call config function
    config_func = CLOUD_SERVICES[service]
    config_func(**settings)



import boto3
from pathlib import Path
from logging import getLogger
from typing import Union, IO, AnyStr

from daman.services import Provider, Progress


logger = getLogger(__name__)


class AWSProvider(Provider):
    def __init__(self, config):
        self.bucket = config["service"]["name"]
        self.s3 = boto3.resource("s3")

    def download(
        self, key: str, file_path: Union[str, Path] = None, buffer: IO[AnyStr] = None
    ):
        # Checking file exists
        msg = f"{key} does not exist in `{self.bucket}` S3 bucket."
        assert key in self.keys, msg

        # Initialising progress bar
        pbar = Progress(size=self.file_size(key), key=key)

        # Download
        if file_path is not None:
            self.s3.meta.client.download_file(
                Bucket=self.bucket, Key=key, Filename=str(file_path), Callback=pbar,
            )
        elif buffer is not None:
            self.s3.meta.client.download_fileobj(
                Fileobj=buffer, Key=key, Bucket=self.bucket, Callback=pbar,
            )
        else:
            raise ValueError("`buffer` or `file_path` cannot be `None` simultaneously.")

        # close progress bar
        pbar.close()

    def upload(
        self, key: str, file_path: Union[str, Path] = None, buffer: IO[AnyStr] = None
    ):
        if file_path is not None:
            msg = f"{file_path} does not exist."
            assert file_path.exists(), msg
            # Initialising progress bar
            pbar = Progress(size=20, key=key)

            # Upload to bucket
            self.s3.meta.client.upload_file(
                Bucket=self.bucket,
                Key=key,
                Filename=str(file_path),
                Callback=pbar,
                ExtraArgs={"Metadata": {"md5": self.md5(file_path=file_path)}},
            )

            # close progres bar
            pbar.close()
        elif buffer is not None:
            # Initialising progress bar
            pbar = Progress(size=buffer.getbuffer().nbytes, key=key)
            # Upload to bucket
            self.s3.meta.client.upload_fileobj(
                Fileobj=buffer,
                Key=key,
                Bucket=self.bucket,
                Callback=pbar,
                ExtraArgs={"Metadata": {"md5": self.md5(buffer=buffer)}},
            )

            # close progres bar
            pbar.close()
        else:
            msg = "either `file_path` or `buffer` must be provided."
            raise ValueError(msg)

    def delete(self, key: str):
        if key in self.keys:
            self.s3.meta.client.delete_object(Bucket=self.bucket, Key=key)
        else:
            logger.warning(f"`{key}` already deleted from cloud service.")

    @property
    def keys(self):
        return [file.key for file in self.s3.Bucket(self.bucket).objects.all()]

    def file_size(self, key: str):
        return self.s3.Bucket(self.bucket).Object(key).content_length

    def check_valid(self, key: str, file_path: Union[str, Path]):
        # get local etag
        msg = f"{file_path} does not exist."
        assert file_path.exists(), msg
        local_md5 = self.md5(file_path=file_path)
        # get remote etag
        msg = f"{key} is not available on `{self.bucket}` S3 bucket."
        if key in self.keys:
            remote_md5 = self.s3.Bucket(self.bucket).Object(key).metadata["md5"]
            return local_md5 == remote_md5
        else:
            logger.warning(msg)
            return True



from daman.commands.configure import configure_command
from daman.commands.download import download_command
from daman.commands.aws import configure_aws_command
from daman.commands.delete import delete_command
from daman.commands.clear import clear_command
from daman.commands.summary import summary_command



import orjson

from daman import CONFIG_DIR


class DataRegistery:
    def __init__(self):
        self.registery_path = CONFIG_DIR / "data_registery.json"

    def read(self):
        if self.registery_path.exists():
            with self.registery_path.open("rb") as fr:
                return orjson.loads(fr.read())
        return {}

    def write(self, registery):
        with self.registery_path.open("wb") as fw:
            fw.write(orjson.dumps(registery))

    def __getitem__(self, key):
        registery = self.read()

        return registery[key]

    def __setitem__(self, key, value):
        registery = self.read()

        registery[key] = value
        self.write(registery)

    def __delitem__(self, key):
        registery = self.read()

        del registery[key]

        self.write(registery)

    @property
    def keys(self):
        return list(self.read().keys())

    @property
    def values(self):
        return list(self.read().values())

    @property
    def items(self):
        return list(self.read().items())

    def __contains__(self, item):
        registery = self.read()
        return item in registery



from pathlib import Path
from logging import getLogger
from configparser import ConfigParser


logger = getLogger(__name__)

HOME = Path().home()
AWS_DIR = HOME / ".aws"


def configure_aws(
    access_key_id: str, secret_access_key: str, region: str = None
) -> None:
    """Short summary.

    Parameters
    ----------
    access_key_id : str
        Description of parameter `access_key_id`.
    secret_access_key : str
        Description of parameter `secret_access_key`.
    region : str
        Description of parameter `region`.

    Returns
    -------
    None
        Description of returned object.

    """
    AWS_DIR.mkdir(parents=True, exist_ok=True)
    logger.info(f"updating AWS credentials: {AWS_DIR / 'credentials'}")

    aws_credentials = ConfigParser()
    if (AWS_DIR / "credentials").exists():
        with (AWS_DIR / "credentials").open("r") as fr:
            aws_credentials.read_file(fr)

    for key, value in [
        ("aws_access_key_id", access_key_id),
        ("aws_secret_access_key", secret_access_key),
    ]:
        set_value(aws_credentials, key, value)

    with (AWS_DIR / "credentials").open("w") as fw:
        aws_credentials.write(fw)

    if region is not None:
        logger.info(f"updating AWS config: {AWS_DIR / 'config'}")
        aws_config = ConfigParser()
        if (AWS_DIR / "config").exists():
            with (AWS_DIR / "config").open() as fr:
                aws_config.read_file(fr)

        set_value(aws_config, "region", region)

        with (AWS_DIR / "config").open("w") as fw:
            aws_config.write(fw)


def set_value(config, key, value):
    """Short summary.

    Parameters
    ----------
    config : type
        Description of parameter `config`.
    key : type
        Description of parameter `key`.
    value : type
        Description of parameter `value`.

    Returns
    -------
    type
        Description of returned object.

    """
    if not config.has_section("default"):
        config.add_section("default")

    if config.has_option("default", value):
        if value != config[f"default_{key}", value]:
            logger.warning(
                f"'{key}' already defined and will be changed from '{config[f'default.{key}']}' to '{value}'"
            )
        else:
            logger.info(f"'{key}' set to '{value}'")

    config.set("default", key, value)



import argparse

from daman.configure import configure_aws


def configure_aws_command():
    parser = argparse.ArgumentParser(description="Sets up AWS service.")
    parser.add_argument(
        "--access_key_id", type=str, help="AWS access_key_id", required=True,
    )
    parser.add_argument(
        "--secret_access_key", type=str, help="AWS secret_access_key", required=True,
    )
    parser.add_argument(
        "--region", type=str, help="AWS region", required=False,
    )
    args = parser.parse_args()

    configure_aws(
        access_key_id=args.access_key_id,
        secret_access_key=args.secret_access_key,
        region=args.region,
    )



import os
import joblib
import psutil
import pandas as pd
from io import BytesIO
from logging import getLogger
from pathlib import Path
from os.path import getsize
from datetime import datetime

from daman.utils import dir_size
from configparser import ConfigParser

from daman.configure import CONFIG_DIR
from daman.services import PROVIDERS
from daman.data import DataRegistery


logger = getLogger(__name__)


class DataManager:
    @property
    def config(self):
        config = ConfigParser()
        config.read(CONFIG_DIR / "config")
        return config

    @property
    def data_folder(self):
        config = self.config
        d_folder = Path(config["local"]["data_dir"])
        if not d_folder.exists():
            d_folder.mkdir(parents=True, exist_ok=True)
        return d_folder

    @property
    def registery(self):
        return DataRegistery()

    @property
    def service(self):
        req_service = self.config["service"]["service"]
        if req_service in PROVIDERS:
            return PROVIDERS[req_service](config=self.config)
        else:
            raise KeyError(
                f"service `{req_service}` requested is not available among provided services."
            )

    def pull(
        self,
        key: str,
        force: bool = False,
        persist: bool = None,
        memory_only: bool = False,
    ):
        msg = f"key '{key}' not found"
        assert key in self.service.keys, msg
        if key in self.registery and not force:
            if persist is not None:
                self.registery[key]["persist"] = persist
            self.registery[key]["used"] += 1
            self.registery[key]["last_used"] = datetime.now()
            file_path = Path(self.registery[key]["path"])
            is_valid = self.service.check_valid(key=key, file_path=file_path)
            if not is_valid:
                logger.warning(
                    f"local '{key}' file doesn't match remote version. Please pull it again using `force`."
                )
            logger.info(f"data `{key}` available locally.")
            obj = joblib.load(file_path)
            return obj["data"], obj["meta"]

        # download file
        if persist is None:
            persist = False

        if memory_only:
            with BytesIO() as buffer:
                buffer.seek(0)
                self.service.download(key=key, buffer=buffer)
                obj = joblib.load(buffer)
                return obj["data"], obj["meta"]
        else:
            logger.info(f"downloading `{key}` file.")
            self.clear_disc(key=key)
            file_path = (self.data_folder / key).resolve()
            self.service.download(key=key, file_path=file_path)
            self.registery[key] = {
                "path": str(file_path),
                "used": 1,
                "created_at": datetime.now(),
                "last_used": datetime.now(),
                "size": getsize(str(file_path)),
                "persist": persist,
            }
            obj = joblib.load(file_path)
            return obj["data"], obj["meta"]

    def push(
        self,
        obj: object,
        key: str,
        meta: object = None,
        local: bool = True,
        force: bool = False,
        persist: bool = False,
    ):
        obj = {"data": obj, "meta": meta}
        if key in self.service.keys:
            if force:
                logger.warning(f"`{key}` already in use and will be over-written.")
            else:
                err_msg = f"`{key}` already in use. Please choose a different key or use --force to enforce key override."
                logger.error(err_msg)
                raise KeyError(err_msg)

        with BytesIO() as file_buffer:
            joblib.dump(obj, file_buffer)
            # reset buffer pointer
            file_buffer.seek(0)

            if local:
                logger.info(f"ensuring disc space available")
                self.clear_disc(space=file_buffer.getbuffer().nbytes / 2 ** 20)

                logger.info(f"storing `{key}` data locally.")
                # store locally
                file_path = (self.data_folder / key).resolve()
                with file_path.open("wb") as fw:
                    fw.write(file_buffer.read())

                logger.info(f"uploading {key} to cloud service.")
                # upload to cloud
                self.service.upload(key=key, file_path=file_path)

                # update registery
                self.registery[key] = {
                    "path": str(file_path),
                    "used": 1,
                    "created_at": datetime.now(),
                    "last_used": datetime.now(),
                    "size": getsize(str(file_path)),
                    "persist": persist,
                }
            else:
                logger.info(f"uploading {key} to cloud service.")
                # upload to cloud
                self.service.upload(key=key, buffer=file_buffer)

    def delete(self, key: str, local: bool = True, remote: bool = False):
        if local:
            logger.info(f"deleting `{key}` data locally.")
            item = self.registery[key]
            # delete file
            del self.registery[key]
            # delete local file
            os.remove(item["path"])
        if remote:
            logger.info(f"deleting `{key}` data on cloud service.")
            # delete remote file
            self.service.delete(key=key)

    @property
    def summary(self):
        service_keys = list(self.service.keys)
        return pd.DataFrame(
            [
                {
                    "key": key,
                    "local": key in self.registery,
                    "remote": key in service_keys,
                    "size (MB)": round(
                        (
                            self.registery[key]["size"]
                            if key in self.registery.keys
                            else self.service.file_size(key)
                        )
                        / 2 ** 20,
                        2,
                    ),
                }
                for key in (set(self.service.keys) | set(self.registery.keys))
            ]
        )

    def available_disc(self):
        if self.config["local"]["allocated_space"] is None:
            free_space = (
                psutil.disk_usage(self.data_folder).free() / 2 ** 20
            )  # disc space in megabytes
        else:
            free_space = int(self.config["local"]["allocated_space"]) - dir_size(
                self.data_folder
            )
        return free_space

    def clear_disc(
        self, key: str = None, space: int = None, ignore_persist: bool = False
    ):
        if key is not None:
            file_size = self.service.file_size(key=key) / 2 ** 20
        else:
            file_size = space

        while self.available_disc() < file_size:
            logger.info(
                f"deleting `{file_size - self.available_disc()}` data on cloud service."
            )
            items = sorted(
                self.registery.items,
                key=lambda x: (x[1]["persist"], x[1]["last_used"], x[1]["used"]),
            )

            if len(items) > 0:
                del_key, del_item = items[0]
                if not del_item["persist"] or ignore_persist:
                    self.delete(key=del_key, local=True, remote=False)
                    continue

            raise IOError(
                f"Not enough space available. Only {self.available_disc()} MB available but {file_size} MB required. No additional file can be deleted."
            )



import argparse
from pathlib import Path

from daman import CONFIG_DIR
from daman.services import PROVIDERS
from daman.configure import configure


def configure_command():
    parser = argparse.ArgumentParser(description="Sets up daman package.")
    parser.add_argument(
        "--storage_name",
        type=str,
        help="name of the storage to push to and pull from",
        required=True,
    )
    parser.add_argument(
        "--service",
        type=str,
        help="type of cloud provider used",
        required=True,
        choices=list(PROVIDERS.keys()),
    )
    parser.add_argument(
        "--local_dir",
        type=Path,
        help="local directory to store data at",
        default=(CONFIG_DIR / "data"),
    )
    parser.add_argument(
        "--allocated_space",
        type=int,
        default=None,
        help="disc space allocated to local directory",
    )
    args = parser.parse_args()

    configure(
        storage_name=args.storage_name,
        service=args.service,
        local_dir=args.local_dir,
        allocated_space=args.allocated_space,
    )
